/* Aranha Module loading interface
 *
 * Copyright Daniel Silverstone <dsilvers@digital-scurf.org>
 * Copyright (C) 2004-2005 Pepperfish Ltd
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "config.h"
#include "modules.h"
#include "configurator.h"
#include "lua.h"
#include "lauxlib.h"

#include <dlfcn.h>

lua_State *ML = 0; /* What sits at the top of ML's stack is, a table */
int module_system_forked = 0; /* Have we forked yet? */

#define GET_DLTABLE lua_settop(ML,0); lua_pushlightuserdata(ML, ML); lua_gettable(ML, LUA_REGISTRYINDEX)

static void *
aux_aranha_dlopen( const char* fname )
{
  if(!ML)return (void*)0;
  GET_DLTABLE;
  lua_pushstring(ML,fname);
  lua_gettable(ML,1);
  if (lua_isnil(ML,2)) {
    lua_settop(ML,1);
    lua_pushstring(ML, fname);
    lua_pushlightuserdata(ML, dlopen(fname, RTLD_LOCAL | RTLD_NOW));
    if( lua_touserdata(ML,-1) == (void*)0 ) {
      /* It was a null, remove that and push the error string instead */
      lua_settop(ML,2);
      lua_pushstring(ML, dlerror());
    } else {
      void *symbol = dlsym(lua_touserdata(ML,-1), "aranha_module_loaded");
      if( symbol ) {
        lua_pushcclosure(ML, *(lua_CFunction*)&symbol, 0);
        lua_pushboolean(ML, module_system_forked);
        lua_pcall(ML, 1, 0, 0);
      }
    }
    lua_settable(ML,1);
    return aux_aranha_dlopen(fname);
  } else {
    return lua_touserdata(ML,2);
  }
}

static const char*
aux_aranha_get_reason(const char* fname)
{
  GET_DLTABLE;
  lua_pushstring(ML, fname);
  lua_gettable(ML,1);
  return lua_tostring(ML,2);
}

void aranha_dlclose_everything()
{
  if(!ML)return;
  invoke_module_event(ML, preunload);
  GET_DLTABLE;
  lua_pushnil(ML);
  while( lua_next(ML,1) != 0 ) {
    void *handle = lua_touserdata(ML,-1);
    if( handle ) dlclose(handle);
    lua_settop(ML,2);
  }
  lua_close(ML);
  ML = 0;
}

static int
L_dlsym(lua_State *L)
{
  const char* libfile = luaL_checkstring(L, 1);
  const char* symbname = luaL_checkstring(L, 2);
  void *handle = aux_aranha_dlopen(libfile);
  void *symbol = handle ? dlsym(handle,symbname) : 0;
  
  lua_settop(L,3);
  
  if( !handle ) {
    lua_pushstring(L, aux_aranha_get_reason(libfile));
    lua_error(L);
  }
  if( !symbol ) {
    lua_pushstring(L, dlerror());
    lua_error(L);
  }
  /* Otherwise, joy of joys, we can return a CFunction on the stack */
  lua_pushlightuserdata(L, ML);
  lua_pushcclosure(L, *(lua_CFunction*)&symbol, 1);
  /* If we were provided with a table we use it as the fenv. */
  if( lua_istable(L, 3) ) {
    lua_pushvalue(L,3);
    lua_setfenv(L,-2);
  }
  return 1;
}

void 
open_module_loader(lua_State *L)
{
  if (!ML) {
    ML = lua_open();
    lua_settop(ML,0); 
    lua_pushlightuserdata(ML, ML);
    lua_newtable(ML);
    lua_settable(ML, LUA_REGISTRYINDEX);
  }
  lua_pushcclosure(L, L_dlsym, 0);
  lua_setglobal(L, "__dlsym");
}

void
_invoke_module_event(lua_State *L, const char* event)
{
  if(!ML)return;
  GET_DLTABLE;
  lua_pushnil(ML);
  while( lua_next(ML,1) != 0 ) {
    void *handle = lua_touserdata(ML,-1);
    void *symbol = handle ? dlsym(handle, event) : 0;
    if( symbol ) {
      lua_pushlightuserdata(L,ML);
      lua_pushcclosure(L, *(lua_CFunction*)&symbol, 1);
      lua_pcall(L, 0, 0, 0);
    }
    lua_settop(ML,2);
  }
}
