/* Aranha FastCGI handler code
 * 
 * Copyright Daniel Silverstone <dsilvers@digital-scurf.org>
 * Copyright (C) 2004-2005 Pepperfish Ltd
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "config.h"

#include "lua_fastcgi.h"

#if !HAVE_STRING_H || !HAVE_STDLIB_H
#error Need string.h and stdlib.h
#endif

#include <string.h>
#include <stdlib.h>

struct {
  FCGX_Stream *in, *out, *err;
} CurrentState;

static void L_PutBytes( lua_State *L, FCGX_Stream *stream )
{
  /* Write all the entries from the lua stack into stream */
  /* On success, return 1, on failure return nil [ on lua stack ] */
  /* Always empties the lua stack and puts a single return onto it */
  int topofstack = lua_gettop(L);
  int index = 0;
  const char* c;
  int allok = 1;
  
  while( ++index <= topofstack ) {
    switch( lua_type( L, index ) ) {
    case LUA_TSTRING:
    case LUA_TNUMBER:
      c = lua_tostring( L, index );
      if( FCGX_PutStr( c, lua_strlen( L, index ), stream ) < 0 )
        allok = 0;
      break;
    case LUA_TNIL:
      if( FCGX_PutStr( "(nil)", 5, stream ) < 0 )
        allok = 0;
      break;
    case LUA_TFUNCTION:
      if( FCGX_PutStr( "(function)", 10, stream ) < 0 )
        allok = 0;
      break;
    default:
      if( FCGX_PutStr( "(?unknown?)", 11, stream ) < 0 )
        allok = 0;
    }
    if( allok == 0 ) break;
  }
  
  lua_settop( L, 0 );
  
  if( allok )
    lua_pushnumber( L, 1 );
  else
    lua_pushnil( L );
}

int L_fcgi_out( lua_State *L )
{
  if( CurrentState.out )
    L_PutBytes( L, CurrentState.out );
  else {
    lua_settop( L, 0 );
    lua_pushnil( L );
  }
  return 1;
}

int L_fcgi_err( lua_State *L )
{
  if( CurrentState.err )
    L_PutBytes( L, CurrentState.err );
  else {
    lua_settop( L, 0 );
    lua_pushnil( L );
  }
  return 1;
}

int L_fcgi_in( lua_State *L )
{
  /* Purpose, read as many bytes as possible from the input stream */
  /* Optionally limited by an argument on the stack */
  int max = 0;
  int bytesread = 0;
  static char buffer[16384];
  
  if( lua_gettop( L ) > 1 ) {
    lua_pushstring( L, "FastCGI.Input takes zero or one argument only" );
    lua_error(L);
  }
  if( lua_gettop( L ) == 1 ) {
    if( lua_isnumber( L, 1 ) == 0 ) {
      lua_pushstring( L, "FastCGI.Input takes a numeric argument only" );
      lua_error(L);
    }
    max = lua_tonumber( L, 1 );
  }
  
  lua_settop( L, 0 );
  
  if( CurrentState.in == 0 ) {
    lua_pushnil( L );
    return 1; /* No point in trying if there's no stream */
  }
  
  while( 1 ) {
    int want2read = 
      (max == 0) ? 16384 :
      (((max-bytesread)>16384) ? 16384 : (max-bytesread));
    int blockread;
    if( want2read == 0 ) break;
    blockread = FCGX_GetStr( buffer, want2read, CurrentState.in );
    if( blockread > 0 ) {
      lua_pushlstring( L, buffer, blockread );
      bytesread += blockread;
    }
    if( blockread < want2read ) break;
    if( lua_gettop( L ) > 1 )
      lua_concat( L, 2 );
  }
  
  if( lua_gettop( L ) > 1 )
    lua_concat( L, 2 );
  
  if( bytesread == 0 ) {
    lua_settop( L, 0 );
    lua_pushnil( L );
  }
  
  return 1;  
}

static int L_fcgi_flush( lua_State *L ) {
  if( CurrentState.out )
    FCGX_FFlush( CurrentState.out );
  if( CurrentState.err )
    FCGX_FFlush( CurrentState.err );
  return 0;
}

static void atexit_handler()
{
  if( CurrentState.in )
    FCGX_Finish();
}

int InitialiseFastCGILibrary( lua_State* L )
{
  /* Initialising the FastCGI Library involves (fortunately)
     very little work
  */
  CurrentState.in = CurrentState.out = CurrentState.err = 0;
  
  /* Theoretically L should have no elements on the stack */
  
  if( lua_gettop( L ) > 0 )
    return -1;
  
  lua_newtable( L ); /* Index 1 -- Will be global 'FastCGI' table */
  
  lua_pushlstring( L, "Output", 6 );
  lua_pushcclosure( L, L_fcgi_out, 0 );
  lua_settable( L, 1 );
    
  lua_pushlstring( L, "Error", 5 );
  lua_pushcclosure( L, L_fcgi_err, 0 );
  lua_settable( L, 1 );
    
  lua_pushlstring( L, "Input", 5 );
  lua_pushcclosure( L, L_fcgi_in, 0 );
  lua_settable( L, 1 );
  
  lua_pushlstring( L, "Flush", 5 );
  lua_pushcclosure( L, L_fcgi_flush, 0 );
  lua_settable( L, 1 );
  
  lua_setglobal( L, "FastCGI" );
 
  atexit( &atexit_handler );
  
  return 0;
}

int AcceptFCGIRequest( lua_State *L )
{
  /* Purpose here is to accept a request and manipulate the FastCGI Table */
  FCGX_ParamArray env;
  int p,l;
  
  if( lua_gettop( L ) > 0 ) return -1; /* Not happy with anything stacked */
  
  if( FCGX_Accept( &(CurrentState.in),
                   &(CurrentState.out),
                   &(CurrentState.err),
                   &env ) ) {
    /* We have a problem */
    return -1;
  }
  /* The 'Accept' Ran okay, so let's parse our way through the info
   * we now have about the request
   */
  
  /* The environment */
  lua_getglobal( L, "FastCGI" );
  lua_pushlstring( L, "env", 3 );
  lua_newtable( L ); /* 3 */
  while( *env ) {
    l = strlen( *env );
    p = (strchr( *env, '=' )) - (*env);
    
    lua_pushlstring( L, *env, p );
    lua_pushlstring( L, *env + p + 1, l - (p + 1) );
    lua_settable( L, 3 ); /* env[key] = value */
    ++env;
  }
  lua_settable(L, 1 ); /* FastCGI[env] = envtable */
  
  /* All done really */
  lua_settop( L, 0 );
  
  return 0;
}

void TerminateFCGIRequest()
{
  FCGX_Finish();
  CurrentState.in = CurrentState.out = CurrentState.err = 0;
}
