/*
** $Id:
** Code Utilities
** See Copyright Notice in lua.h
*/

#define lcutils_c
#define LUA_CORE

#include <stdlib.h>

#include "lua.h"
#include "lcode.h"
#include "lcutils.h"
#include "llex.h"
#include "lopcodes.h"
#include "lparser.h"



/* 
 * This function scans a range of opcodes, and corrects all register
 * references >= base by adding offset to them.
 * 
 * It would be nice if there were a bit in the opcode info which
 * specified whether or not A is a register; since there isn't, I
 * inserted manual checks for the specific ops. I think that's ugly,
 * but as far as I can see there are only four ops which don't use
 * A as a register. However, the code will incorrectly relocate
 * the upvalue instructions following OP_CLOSURE if it is relocating
 * from register 0. Possible solutions compiling NO_REG into these
 * instructions and/or using different opcodes. However, I don't
 * think it's actually a problem.
 *
 * Most of the code for checking hwm is copied from ldebug.c, although
 * it is slightly simplified in a few cases (I don't bother checking
 * for open instructions, for example, since the checks do no harm.)
 *
 * I think the clean solution here is to put two functions in lopcodes.c:
 * int luaP_maxreach(Instruction i) 
 *   returns the largest stack slot accessed by instruction i
 * int luaP_modifies(Instruction i, int s)
 *   returns true if instruction i might modify stack slot s
 * symbcheck requires both these services; relocate only requires the former.
 * This would keep everything in one place, which I think is a cleaner
 * solution, even though it would be slightly less efficient. I don't think
 * either this code or the code in ldebug.c is performance-critical.
 * 
 * (Had it been critical, I might have skimped on some of the tests below.
 * Not all of them are necessary; in many cases, we can actually know for
 * certain that the limit will be the target of some other instruction.
 * But it saved thought putting them all in.)
 */


void luaCU_shiftcode (FuncState *fs, int frompc, int topc,
                             int base, int offset) { 
  int pc;
  Instruction *code = fs->f->code;
  int hwm = base;
  lua_assert(base >= 2);
  for (pc = frompc; pc < topc; ++pc) {
    Instruction i = code[pc];
    OpCode op = GET_OPCODE(i);
    int a = GETARG_A(i);
    int b = GETARG_B(i);
    int c = GETARG_C(i);
    switch(op) {
      case OP_SELF:
      case OP_NEWTABLE:
        if (hwm < a + 1) hwm = a + 1; 
        break;
      case OP_MOVE3:
        if (hwm < a + 2) hwm = a + 2;
        /* if (hwm < b + 2) hwm = b + 2; * b+3 <= a */
        break;
      case OP_TAILCTRCLOSE:
      case OP_CTRCLOSE:
        if (hwm < a + c - 1) hwm = a + c - 1;
        if (hwm < a + b + LUA_CTRARGS) hwm = a + b + LUA_CTRARGS;
        break;
      case OP_CTROPEN:
        if (hwm < a + b + 2) hwm = a + b + 2;
        break;
      case OP_TFORCALL:
        /* if (hwm < b + 2) hwm = b + 2;  * b+3 <= a */
        if (hwm < a + c - 1) hwm = a + c - 1;
        break;
      case OP_CALL:
      case OP_TAILCALL:
        if (hwm < a + b - 1) hwm = a + b - 1;
        if (hwm < a + c - 1) hwm = a + c - 1;
        break;
      case OP_RETURN:
        if (hwm < a + b - 1) hwm = a + b - 1;
        break;
      case OP_SETLIST:
        if (hwm < c + b - 1) hwm = c + b - 1;
        break;
      case OP_VARARG:
        if (hwm < a + b - 2) hwm = a + b - 2;
        break;
      default:
        break;
    }
    if (getOpMode(op) == iABC) {
      if (op < OP_EQ || op > OP_LE) {
        if (a >= base && a != NO_REG) {
          if (hwm < a) hwm = a;
          a += offset;
        }
      }
      if (getBMode(op) >= OpArgR) {
        if (!ISK(b) && b >= base) {
          if (hwm < b) hwm = b;
          b += offset;
        }
      }
      if (getCMode(op) >= OpArgR) {
        if (!ISK(c) && c >= base) {
          if (hwm < c) hwm = c;
          c += offset;
        }
      }
      code[pc] = CREATE_ABC(op, a, b, c);
    }
    else if (a >= base && op != OP_JMP) {
      if (hwm < a) hwm = a;
      SETARG_A(i, a + offset);
      code[pc] = i;
    }
  }
  luaK_checkstack(fs, hwm + offset + 1);
}


/* Rotate 'r' instructions starting at frompc to the end of the code. */

void luaCU_rotatecode (const FuncState *fs, int frompc, int r) {
  int n = fs->pc - frompc;
  if (r > 0 && r < n) {
    Instruction *code = &fs->f->code[frompc];
    int *lineinfo = &fs->f->lineinfo[frompc];
    short *scopeinfo = &fs->f->scopeinfo[frompc];
    int i = n;
    int j = n;
    /* Fix up relative branches. This code could be improved */
    for (--j; j >= r; --j) {
      Instruction insn = code[j];
      if (getOpMode(GET_OPCODE(insn)) == iAsBx) {
        int offset = GETARG_sBx(insn);
        int targ = offset + j + 1;
        if (targ < 0 || targ >= n) offset += r;
        else if (targ < r) offset += n;
        else continue;
        if (abs(offset) > MAXARG_sBx) 
          luaX_syntaxerror(fs->ls, "control structure too long");
        SETARG_sBx(code[j], offset);
      }
    }
    for (; j >= 0; --j) {
      Instruction insn = code[j];
      if (getOpMode(GET_OPCODE(insn)) == iAsBx) {
        int offset = GETARG_sBx(insn);
        int targ = offset + j + 1;
        if (targ < 0 || targ >= n) offset -= n - r;
        else if (targ >= r) offset -= n;
        else continue;
        if (abs(offset) > MAXARG_sBx) 
          luaX_syntaxerror(fs->ls, "control structure too long");
        SETARG_sBx(code[j], offset);
      }
    }
    /* Now rotate the code */
    j = 0;
    do {
      Instruction ctemp = code[j];
      int ltemp = lineinfo[j];
      short stemp = scopeinfo[j];
      int k = j;
      int kk = j + r;
      do {
        --i;
        code[k] = code[kk];
        lineinfo[k] = lineinfo[kk];
        scopeinfo[k] = scopeinfo[kk];
        k = kk;
        kk += r;
        if (kk >= n) kk -= n;
      } while (kk != j);
      code[k] = ctemp;
      lineinfo[k] = ltemp;
      scopeinfo[k] = stemp;
      ++j;
    } while (i > j);
  }
}
