/*
** $Id: loslib.c,v 1.19 2006/04/26 18:19:49 roberto Exp $
** Standard Operating System library
** See Copyright Notice in lua.h
*/


#include <errno.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#define loslib_c
#define LUA_LIB

#include "lua.h"

#include "lauxlib.h"
#include "lualib.h"


static int os_pushresult (lua_State *L, int i, const char *filename) {
  int en = errno;  /* calls to Lua API may change this value */
  if (i) {
    lua_pushboolean(L, 1);
    return 1;
  }
  else {
    lua_pushnil(L);
    lua_pushfstring(L, "%s: %s", filename, strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
}


static int os_execute (lua_State *L) {
  lua_pushinteger(L, system(luaL_optstring(L, 1, NULL)));
  return 1;
}


static int os_remove (lua_State *L) {
  const char *filename = luaL_checkstring(L, 1);
  return os_pushresult(L, remove(filename) == 0, filename);
}


static int os_rename (lua_State *L) {
  const char *fromname = luaL_checkstring(L, 1);
  const char *toname = luaL_checkstring(L, 2);
  return os_pushresult(L, rename(fromname, toname) == 0, fromname);
}

static int internal_fetch_stat_info(lua_State *L, struct stat *s, int link) {
  const char *p = luaL_checkstring(L,1);
  if( link )
    return lstat(p, s);
  return stat(p,s);
}

static int os_exists(lua_State *L) {
  static struct stat s;
  int r = internal_fetch_stat_info( L, &s, 0 );
  lua_settop(L,0);
  lua_pushboolean(L, r==0);
  return 1;
}

static int os_isdir(lua_State *L) {
  static struct stat s;
  int r = internal_fetch_stat_info( L, &s, 0 );
  if( r != 0 ) return 0;
  lua_settop(L,0);
  lua_pushboolean(L, S_ISDIR(s.st_mode));
  return 1;
}

static int os_islink(lua_State *L) {
  static struct stat s;
  int r = internal_fetch_stat_info( L, &s, 1 );
  if( r != 0 ) return 0;
  lua_settop(L,0);
  lua_pushboolean(L, S_ISLNK(s.st_mode));
  return 1;
}

static int os_fileinfo(lua_State *L) {
  const char *p = luaL_checkstring(L,1);
  int link = luaL_optboolean(L, 2, 0);
  static struct stat s;
  if( link ) {
    if( lstat(p, &s) != 0 ) return 0;
  } else {
    if( stat(p, &s) != 0 ) return 0;
  }
  lua_settop(L,0);
  lua_newtable(L);
#define PUSHFIELD(N,E) lua_pushliteral(L,#N); lua_pushnumber(L,s.st_##E); lua_settable(L,1)
  PUSHFIELD(device, dev);
  PUSHFIELD(inode, ino);
  PUSHFIELD(linkcount, nlink);
  PUSHFIELD(user, uid);
  PUSHFIELD(group, gid);
  PUSHFIELD(size, size);
  PUSHFIELD(accesstime, atime);
  PUSHFIELD(modifiedtime, mtime);
  PUSHFIELD(changetime, ctime);
#undef PUSHFIELD
#define PUSHCOND(N,C) lua_pushliteral(L,#N); lua_pushboolean(L,(C)); lua_settable(L,1)
  PUSHCOND(regular,S_ISREG(s.st_mode));
  PUSHCOND(directory,S_ISDIR(s.st_mode));
  PUSHCOND(chardev,S_ISCHR(s.st_mode));
  PUSHCOND(blockdev,S_ISBLK(s.st_mode));
  PUSHCOND(fifo,S_ISFIFO(s.st_mode));
  PUSHCOND(symlink,S_ISLNK(s.st_mode));
  PUSHCOND(socket,S_ISSOCK(s.st_mode));
#undef PUSHCOND
#define PUSHMODE(N,C) lua_pushliteral(L,#N); lua_pushboolean(L,s.st_mode & C); lua_settable(L,-3)
  lua_pushliteral(L,"mode");
  lua_newtable(L);
  PUSHMODE(setuid,S_ISUID);
  PUSHMODE(setgid,S_ISGID);
  PUSHMODE(sticky,S_ISVTX);
  lua_pushliteral(L,"owner");
  lua_newtable(L);
  PUSHMODE(read,    S_IRUSR);
  PUSHMODE(write,   S_IWUSR);
  PUSHMODE(execute, S_IXUSR);
  lua_settable(L,-3);
  lua_pushliteral(L,"group");
  lua_newtable(L);
  PUSHMODE(read,    S_IRGRP);
  PUSHMODE(write,   S_IWGRP);
  PUSHMODE(execute, S_IXGRP);
  lua_settable(L,-3);
  lua_pushliteral(L,"other");
  lua_newtable(L);
  PUSHMODE(read,    S_IROTH);
  PUSHMODE(write,   S_IWOTH);
  PUSHMODE(execute, S_IXOTH);
  lua_settable(L,-3);
  lua_settable(L,-3);
  return 1;
}


static int os_tmpname (lua_State *L) {
  char buff[LUA_TMPNAMBUFSIZE];
  int err;
  lua_tmpnam(buff, err);
  if (err)
    return luaL_error(L, "unable to generate a unique filename");
  lua_pushstring(L, buff);
  return 1;
}


static int os_getenv (lua_State *L) {
  lua_pushstring(L, getenv(luaL_checkstring(L, 1)));  /* if NULL push nil */
  return 1;
}


static int os_clock (lua_State *L) {
  lua_pushnumber(L, ((lua_Number)clock())/(lua_Number)CLOCKS_PER_SEC);
  return 1;
}


/*
** {======================================================
** Time/Date operations
** { year=%Y, month=%m, day=%d, hour=%H, min=%M, sec=%S,
**   wday=%w+1, yday=%j, isdst=? }
** =======================================================
*/

static void setfield (lua_State *L, const char *key, int value) {
  lua_pushinteger(L, value);
  lua_setfield(L, -2, key);
}

static void setboolfield (lua_State *L, const char *key, int value) {
  if (value < 0)  /* undefined? */
    return;  /* does not set field */
  lua_pushboolean(L, value);
  lua_setfield(L, -2, key);
}

static int getboolfield (lua_State *L, const char *key) {
  int res;
  lua_getfield(L, -1, key);
  res = lua_isnil(L, -1) ? -1 : lua_toboolean(L, -1);
  lua_pop(L, 1);
  return res;
}


static int getfield (lua_State *L, const char *key, int d) {
  int res;
  lua_getfield(L, -1, key);
  if (lua_isnumber(L, -1))
    res = (int)lua_tointeger(L, -1);
  else {
    if (d < 0)
      return luaL_error(L, "field " LUA_QS " missing in date table", key);
    res = d;
  }
  lua_pop(L, 1);
  return res;
}


static int os_date (lua_State *L) {
  const char *s = luaL_optstring(L, 1, "%c");
  time_t t = luaL_opt(L, (time_t)luaL_checknumber, 2, time(NULL));
  struct tm *stm;
  if (*s == '!') {  /* UTC? */
    stm = gmtime(&t);
    s++;  /* skip `!' */
  }
  else
    stm = localtime(&t);
  if (stm == NULL)  /* invalid date? */
    lua_pushnil(L);
  else if (strcmp(s, "*t") == 0) {
    lua_createtable(L, 0, 9);  /* 9 = number of fields */
    setfield(L, "sec", stm->tm_sec);
    setfield(L, "min", stm->tm_min);
    setfield(L, "hour", stm->tm_hour);
    setfield(L, "day", stm->tm_mday);
    setfield(L, "month", stm->tm_mon+1);
    setfield(L, "year", stm->tm_year+1900);
    setfield(L, "wday", stm->tm_wday+1);
    setfield(L, "yday", stm->tm_yday+1);
    setboolfield(L, "isdst", stm->tm_isdst);
  }
  else {
    char b[256];
    if (strftime(b, sizeof(b), s, stm))
      lua_pushstring(L, b);
    else
      return luaL_error(L, LUA_QL("date") " format too long");
  }
  return 1;
}


static int os_time (lua_State *L) {
  time_t t;
  if (lua_isnoneornil(L, 1))  /* called without args? */
    t = time(NULL);  /* get current time */
  else {
    struct tm ts;
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_settop(L, 1);  /* make sure table is at the top */
    ts.tm_sec = getfield(L, "sec", 0);
    ts.tm_min = getfield(L, "min", 0);
    ts.tm_hour = getfield(L, "hour", 12);
    ts.tm_mday = getfield(L, "day", -1);
    ts.tm_mon = getfield(L, "month", -1) - 1;
    ts.tm_year = getfield(L, "year", -1) - 1900;
    ts.tm_isdst = getboolfield(L, "isdst");
    t = mktime(&ts);
  }
  if (t == (time_t)(-1))
    lua_pushnil(L);
  else
    lua_pushnumber(L, (lua_Number)t);
  return 1;
}


static int os_difftime (lua_State *L) {
  lua_pushnumber(L, difftime((time_t)(luaL_checknumber(L, 1)),
                             (time_t)(luaL_optnumber(L, 2, 0))));
  return 1;
}

/* }====================================================== */


static int os_setlocale (lua_State *L) {
  static const int cat[] = {LC_ALL, LC_COLLATE, LC_CTYPE, LC_MONETARY,
                      LC_NUMERIC, LC_TIME};
  static const char *const catnames[] = {"all", "collate", "ctype", "monetary",
     "numeric", "time", NULL};
  const char *l = luaL_optstring(L, 1, NULL);
  int op = luaL_checkoption(L, 2, "all", catnames);
  lua_pushstring(L, setlocale(cat[op], l));
  return 1;
}


static int os_exit (lua_State *L) {
  exit(luaL_optint(L, 1, EXIT_SUCCESS));
  return 0;  /* to avoid warnings */
}

static int
os_readdir( lua_State *L )
{
  /* IN: path to dir to read */
  /* OUT: list of leaves in that dir minus . and .. */
  DIR *d;
  struct dirent *de;
  int c = 0;
  const char* dirname = luaL_checkstring(L, 1);
  d = opendir(dirname);
  if( !d ) {
    lua_pushfstring(L, "Unable to open directory %s", d);
    return lua_error(L);
  }
  lua_newtable(L);
  while( (de = readdir(d)) ) {
    if( strcmp(".", de->d_name) == 0 || strcmp("..", de->d_name) == 0) continue;
    lua_pushnumber(L, ++c);
    lua_pushstring(L, de->d_name);
    lua_settable(L, -3);
  }
  closedir(d);
  return 1;
}

static const luaL_Reg syslib[] = {
  {"clock",     os_clock},
  {"date",      os_date},
  {"difftime",  os_difftime},
  {"execute",   os_execute},
  {"exit",      os_exit},
  {"getenv",    os_getenv},
  {"readdir",   os_readdir},
  {"remove",    os_remove},
  {"rename",    os_rename},
  {"setlocale", os_setlocale},
  {"time",      os_time},
  {"exists",    os_exists},
  {"isdir",     os_isdir},
  {"islink",    os_islink},
  {"fileinfo",  os_fileinfo},
  {"tmpname",   os_tmpname},
  {NULL, NULL}
};

/* }====================================================== */



LUALIB_API int luaopen_os (lua_State *L) {
  luaL_register(L, LUA_OSLIBNAME, syslib);
  return 1;
}

