/*
** $Id: lparser.c,v 2.42 2006/06/05 15:57:59 roberto Exp $
** Lua Parser
** See Copyright Notice in lua.h
*/


#include <string.h>
#include <stdio.h>

#define lparser_c
#define LUA_CORE

#include "lua.h"

#include "lcode.h"
#include "lcutils.h"
#include "ldebug.h"
#include "ldo.h"
#include "lfunc.h"
#include "llex.h"
#include "lmem.h"
#include "lobject.h"
#include "lopcodes.h"
#include "lparser.h"
#include "lstate.h"
#include "lstring.h"
#include "ltable.h"



#define hasmultret(k)		((k) == VCALL || (k) == VVARARG)

#define luaY_checklimit(fs,v,l,m)	if ((v)>(l)) errorlimit(fs,l,m)


/*
** nodes for block list (list of active blocks)
*/
typedef struct BlockCnt {
  struct BlockCnt *previous;  /* chain */
  int breaklist;  /* list of breaks out of this block */
  short scope;  /* scope enclosing block */
  lu_byte basereg;  /* first free register in block */
  lu_byte upval;  /* true if some variable in the block is an upvalue */
  /* I'd like to ditch the next member, but I'm not sure if I can */
  lu_byte isbreakable;  /* true if `block' is a loop */
} BlockCnt;

static void forclause (LexState *ls, int line, int breakable,
                       void(*bodyfunc)(LexState *ls, void *ud), void *ud);

/*
** prototypes for recursive non-terminal functions
*/
static void chunk (LexState *ls);
static void expr (LexState *ls, expdesc *v);


static void anchor_token (LexState *ls) {
  if (ls->t.token == TK_NAME || ls->t.token == TK_STRING) {
    TString *ts = ls->t.seminfo.ts;
    luaX_newstring(ls, getstr(ts), ts->tsv.len);
  }
}


static void error_expected (LexState *ls, int token) {
  luaX_syntaxerror(ls,
      luaO_pushfstring(ls->L, LUA_QS " expected", luaX_token2str(ls, token)));
}


static void errorlimit (FuncState *fs, int limit, const char *what) {
  const char *msg = (fs->f->linedefined == 0) ?
    luaO_pushfstring(fs->L, "main function has more than %d %s", limit, what) :
    luaO_pushfstring(fs->L, "function at line %d has more than %d %s",
                            fs->f->linedefined, limit, what);
  luaX_lexerror(fs->ls, msg, 0);
}


static int testnext (LexState *ls, int c) {
  if (ls->t.token == c) {
    luaX_next(ls);
    return 1;
  }
  else return 0;
}


static void check (LexState *ls, int c) {
  if (ls->t.token != c)
    error_expected(ls, c);
}

static void checknext (LexState *ls, int c) {
  check(ls, c);
  luaX_next(ls);
}

static int testnextsemi (LexState *ls, const char *s) {
  if (ls->t.token == TK_NAME &&
      strcmp(getstr(ls->t.seminfo.ts), s) == 0) {
    luaX_next(ls);
    return 1;
  }
  else return 0;
}


static void checknextsemi (LexState *ls, const char *s) {
  if (!testnextsemi(ls, s)) {
    luaX_syntaxerror(ls,
        luaO_pushfstring(ls->L, LUA_QS " expected", s));
  }
}


#define check_condition(ls,c,msg) { if (!(c)) luaX_syntaxerror(ls, msg); }


static void check_match (LexState *ls, int what, int who, int where) {
  if (!testnext(ls, what)) {
    if (where == ls->linenumber)
      error_expected(ls, what);
    else {
      luaX_syntaxerror(ls, luaO_pushfstring(ls->L,
             LUA_QS " expected (to close " LUA_QS " at line %d)",
              luaX_token2str(ls, what), luaX_token2str(ls, who), where));
    }
  }
}


static TString *str_checkname (LexState *ls) {
  TString *ts;
  check(ls, TK_NAME);
  ts = ls->t.seminfo.ts;
  luaX_next(ls);
  return ts;
}


static void init_exp (expdesc *e, expkind k, int i) {
  e->f = e->t = NO_JUMP;
  e->k = k;
  e->u.s.info = i;
}


static void codestring (LexState *ls, expdesc *e, TString *s) {
  init_exp(e, VK, luaK_stringK(ls->fs, s));
}


static void checkname(LexState *ls, expdesc *e) {
  codestring(ls, e, str_checkname(ls));
}


#define getscope(fs, s)        (&(fs)->f->scopes[s])
#define scopesize(fs, s)       (getscope(fs, s)->nvars)
#define scopebase(fs, s)       (getscope(fs, s)->slot)
#define scopeend(fs, s)        (getscope(fs, s)->slot + getscope(fs, s)->nvars)


static void registerlocalvar (FuncState *fs, TString *varname) {
  Proto *f = fs->f;
  int oldsize = f->sizelocvars;
  luaM_growvector(fs->L, f->locvars, fs->nlocvars, f->sizelocvars,
                  TString *, SHRT_MAX, "too many local variables");
  while (oldsize < f->sizelocvars) f->locvars[oldsize++] = NULL;
  f->locvars[fs->nlocvars] = varname;
  luaC_objbarrier(fs->L, f, varname);
  fs->nlocvars++;
}


#define new_localvarliteral(fs,s,v) \
  new_localvar(fs, s, luaX_newstring((fs)->ls, "" v, \
               (sizeof(v)/sizeof(char))-1))

  
static void new_localvar (FuncState *fs, int s, TString *name) {
  int n = scopesize(fs, s) + 1;
  luaY_checklimit(fs, scopebase(fs, s) + n, LUAI_MAXVARS, "local variables");
  registerlocalvar(fs, name);
  scopesize(fs, s) = n;
}


static int newscope (FuncState *fs) {
  Proto *f = fs->f;
  Scope *scope;
  luaM_growvector(fs->L, f->scopes, fs->nscopes, f->sizescopes,
      Scope, SHRT_MAX, "too many scopes");
  scope = getscope(fs, fs->nscopes);
  scope->outer = -1;
  scope->locvar = fs->nlocvars;
  scope->nvars = 0;
  scope->slot = fs->freereg;
  return fs->nscopes++;
}


static void reanchorscope (FuncState *fs, int s) {
  getscope(fs, s)->slot = fs->freereg;
}


static int isparentscope (FuncState *fs, int s, int toscope) {
  while (s != toscope && s != -1) s = getscope(fs, s)->outer;
  return s == toscope;
}
  

static void reactivatescope (FuncState *fs, int s) {
  if (s >= 0) {
    int tempreg = scopeend(fs, s);
     lua_assert(isparentscope(fs, s, fs->scope));
    fs->scope = s;
    luaK_reserveregs(fs, tempreg - fs->freereg);
    fs->tempreg = tempreg;
    /* printf("Activating scope %d, tempreg = %d\n", s, fs->tempreg); */
  }
}


static void activatescope (FuncState *fs, int s) {
  /* The scope has not previously been activated */
  lua_assert(getscope(fs, s)->outer == -1);
  getscope(fs, s)->outer = fs->scope;
  reactivatescope(fs, s);
}


static int closescopes (FuncState *fs, int toscope) {
  int s = fs->scope;
  lua_assert(isparentscope(fs, s, toscope));
  fs->scope = toscope;
  fs->freereg = fs->tempreg = toscope >= 0 ? scopeend(fs, toscope) : 0;
  /* printf("Closing to scope %d, tempreg = %d\n", toscope, fs->tempreg); */
  return s;
}


static int indexupvalue (FuncState *fs, TString *name, expdesc *v) {
  int i;
  Proto *f = fs->f;
  int oldsize = f->sizeupvalues;
  for (i=0; i<f->nups; i++) {
    if (fs->upvalues[i].k == v->k && fs->upvalues[i].info == v->u.s.info) {
      lua_assert(f->upvalues[i] == name);
      return i;
    }
  }
  /* new one */
  luaY_checklimit(fs, f->nups + 1, LUAI_MAXUPVALUES, "upvalues");
  luaM_growvector(fs->L, f->upvalues, f->nups, f->sizeupvalues,
                  TString *, MAX_INT, "");
  while (oldsize < f->sizeupvalues) f->upvalues[oldsize++] = NULL;
  f->upvalues[f->nups] = name;
  luaC_objbarrier(fs->L, f, name);
  lua_assert(v->k == VLOCAL || v->k == VUPVAL);
  fs->upvalues[f->nups].k = cast_byte(v->k);
  fs->upvalues[f->nups].info = cast_byte(v->u.s.info);
  return f->nups++;
}


static int searchvar (FuncState *fs, TString *n) {
  int s = fs->scope;
  while (s >= 0) {
    Scope *scope = getscope(fs, s);
    int i;
    for (i = scope->nvars - 1; i >= 0; i--) {
      if (n == fs->f->locvars[scope->locvar + i])
        return scope->slot + i;
    }
    s = scope->outer;
  }
  return -1;
}


static void markupval (FuncState *fs, int level) {
  BlockCnt *bl = fs->bl;
  while (bl && bl->basereg > level) bl = bl->previous;
  if (bl) bl->upval = 1;
}


static int singlevaraux (FuncState *fs, TString *n, expdesc *var, int base) {
  if (fs == NULL) {  /* no more levels? */
    init_exp(var, VGLOBAL, NO_REG);  /* default is global variable */
    return VGLOBAL;
  }
  else {
    int v = searchvar(fs, n);  /* look up at current level */
    if (v >= 0) {
      init_exp(var, VLOCAL, v);
      if (!base)
        markupval(fs, v);  /* local will be used as an upval */
      return VLOCAL;
    }
    else {  /* not found at current level; try upper one */
      if (singlevaraux(fs->prev, n, var, 0) == VGLOBAL)
        return VGLOBAL;
      var->u.s.info = indexupvalue(fs, n, var);  /* else was LOCAL or UPVAL */
      var->k = VUPVAL;  /* upvalue in this level */
      return VUPVAL;
    }
  }
}


static void singlevar (LexState *ls, expdesc *var) {
  TString *varname = str_checkname(ls);
  FuncState *fs = ls->fs;
  if (singlevaraux(fs, varname, var, 1) == VGLOBAL)
    var->u.s.info = luaK_stringK(fs, varname);  /* info points to global name */
}


static void adjust_assign (FuncState *fs, int nvars, int nexps, expdesc *e) {
  int extra = nvars - nexps;
  if (hasmultret(e->k)) {
    extra++;  /* includes call itself */
    if (extra < 0) extra = 0;
    luaK_setreturns(fs, e, extra);  /* last exp. provides the difference */
    if (extra > 1) luaK_reserveregs(fs, extra-1);
  }
  else {
    if (e->k != VVOID) luaK_exp2nextreg(fs, e);  /* close last expression */
    if (extra > 0) {
      int reg = fs->freereg;
      luaK_reserveregs(fs, extra);
      luaK_nil(fs, reg, extra);
    }
  }
}


static void enterlevel (LexState *ls) {
  if (++ls->L->nCcalls > LUAI_MAXCCALLS)
	luaX_lexerror(ls, "chunk has too many syntax levels", 0);
}


#define leavelevel(ls)	((ls)->L->nCcalls--)


static void enterblock (FuncState *fs, BlockCnt *bl, lu_byte isbreakable) {
  bl->breaklist = NO_JUMP;
  bl->isbreakable = isbreakable;
  bl->basereg = fs->freereg;
  bl->scope = fs->scope;
  bl->upval = 0;
  bl->previous = fs->bl;
  fs->bl = bl;
}


static void leaveblock (FuncState *fs) {
  BlockCnt *bl = fs->bl;
  fs->bl = bl->previous;
  closescopes(fs, bl->scope);
  if (bl->upval)
    luaK_codeABC(fs, OP_CLOSE, bl->basereg, 0, 0);
  luaK_patchtohere(fs, bl->breaklist);
  fs->freereg = bl->basereg;  /* free registers */
}


static void mergeblock (FuncState *fs) {
  BlockCnt *bl = fs->bl;
  BlockCnt *prev = bl->previous;
  prev->upval |= bl->upval;
  prev->scope = bl->scope;
  fs->bl = prev;
}


static int needsclose (FuncState *fs, BlockCnt *targ) {
  BlockCnt *bl = fs->bl;
  for (;; bl = bl->previous) {
    lua_assert(bl);
    if (bl->upval) return 1;
    if (bl == targ) return 0;
  }
}


static BlockCnt *findbreakable (FuncState *fs) {
  BlockCnt *bl = fs->bl;
  for (; bl; bl = bl->previous) {
    if (bl->isbreakable) break;
  }
  return bl;
}


static void codebreak (FuncState *fs, BlockCnt *bl) {
  if (needsclose(fs, bl))
    luaK_codeABC(fs, OP_CLOSE, bl->basereg, 0, 0);
  luaK_concat(fs, &bl->breaklist, luaK_jump(fs));
}


static void pushclosure (FuncState *fs, FuncState *func, expdesc *v) {
  Proto *f = fs->f;
  int oldsize = f->sizep;
  int i;
  luaM_growvector(fs->L, f->p, fs->np, f->sizep, Proto *,
                  MAXARG_Bx, "constant table overflow");
  while (oldsize < f->sizep) f->p[oldsize++] = NULL;
  f->p[fs->np++] = func->f;
  luaC_objbarrier(fs->L, f, func->f);
  init_exp(v, VRELOCABLE, luaK_codeABx(fs, OP_CLOSURE, 0, fs->np-1));
  for (i=0; i<func->f->nups; i++) {
    OpCode o = (func->upvalues[i].k == VLOCAL) ? OP_MOVE : OP_GETUPVAL;
    luaK_codeABC(fs, o, 0, func->upvalues[i].info, 0);
  }
}


static void open_func (LexState *ls, FuncState *fs) {
  lua_State *L = ls->L;
  Proto *f = luaF_newproto(L);
  fs->f = f;
  fs->prev = ls->fs;  /* linked list of funcstates */
  fs->ls = ls;
  fs->L = L;
  ls->fs = fs;
  fs->pc = 0;
  fs->lasttarget = -1;
  fs->jpc = NO_JUMP;
  fs->freereg = 0;
  fs->tempreg = 0;
  fs->nk = 0;
  fs->np = 0;
  fs->nlocvars = 0;
  fs->nscopes = 0;
  fs->scope = -1;
  fs->bl = NULL;
  f->source = ls->source;
  f->maxstacksize = 2;  /* registers 0/1 are always valid */
  fs->h = luaH_new(L, 0, 0);
  /* anchor table of constants and prototype (to avoid being collected) */
  sethvalue2s(L, L->top, fs->h);
  incr_top(L);
  setptvalue2s(L, L->top, f);
  incr_top(L);
}


static void close_func (LexState *ls) {
  lua_State *L = ls->L;
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  closescopes(fs, -1);
  luaK_ret(fs, 0, 0);  /* final return */
  luaM_reallocvector(L, f->code, f->sizecode, fs->pc, Instruction);
  f->sizecode = fs->pc;
  luaM_reallocvector(L, f->lineinfo, f->sizelineinfo, fs->pc, int);
  f->sizelineinfo = fs->pc;
  luaM_reallocvector(L, f->scopeinfo, f->sizescopeinfo, fs->pc, short);
  f->sizescopeinfo = fs->pc;
  luaM_reallocvector(L, f->k, f->sizek, fs->nk, TValue);
  f->sizek = fs->nk;
  luaM_reallocvector(L, f->p, f->sizep, fs->np, Proto *);
  f->sizep = fs->np;
  luaM_reallocvector(L, f->locvars, f->sizelocvars, fs->nlocvars, TString *);
  f->sizelocvars = fs->nlocvars;
  luaM_reallocvector(L, f->scopes, f->sizescopes, fs->nscopes, Scope);
  f->sizescopes = fs->nscopes;
  luaM_reallocvector(L, f->upvalues, f->sizeupvalues, f->nups, TString *);
  f->sizeupvalues = f->nups;
  lua_assert(luaG_checkcode(f));
  lua_assert(fs->bl == NULL);
  ls->fs = fs->prev;
  L->top -= 2;  /* remove table and prototype from the stack */
  /* last token read was anchored in defunct function; must reanchor it */
  if (fs) anchor_token(ls);
}


Proto *luaY_parser (lua_State *L, ZIO *z, Mbuffer *buff, Mbuffer *dbuff,
                    const char *name) {
  struct LexState lexstate;
  struct FuncState funcstate;
  lexstate.buff = buff;
  lexstate.dbuff = dbuff;
  luaX_setinput(L, &lexstate, z, luaS_new(L, name));
  open_func(&lexstate, &funcstate);
  funcstate.f->is_vararg = VARARG_ISVARARG;  /* main func. is always vararg */
  luaX_next(&lexstate);  /* read first token */
  if (lexstate.t.token == TK_DOCSTRING) {
    /* Capture this string as the module's documentation */
    funcstate.f->doc = lexstate.t.seminfo.ts;
    luaX_next(&lexstate);
  }
  chunk(&lexstate);
  check(&lexstate, TK_EOS);
  close_func(&lexstate);
  lua_assert(funcstate.prev == NULL);
  lua_assert(funcstate.f->nups == 0);
  lua_assert(lexstate.fs == NULL);
  return funcstate.f;
}



/*============================================================*/
/* GRAMMAR RULES */
/*============================================================*/


static void field (LexState *ls, expdesc *v) {
  /* field -> ['.' | ':'] NAME */
  FuncState *fs = ls->fs;
  expdesc key;
  luaK_exp2anyreg(fs, v);
  luaX_next(ls);  /* skip the dot or colon */
  checkname(ls, &key);
  luaK_indexed(fs, v, &key);
}


static void yindex (LexState *ls, expdesc *v) {
  /* index -> '[' expr ']' */
  luaX_next(ls);  /* skip the '[' */
  expr(ls, v);
  luaK_exp2val(ls->fs, v);
  checknext(ls, ']');
}


/*
** {======================================================================
** Rules for Constructors
** =======================================================================
*/


struct ConsControl {
  expdesc v;  /* last list item read */
  BlockCnt *bl;
  int nh;  /* total number of `record' elements */
  int na;  /* total number of array elements */
  int tostore;  /* number of array elements pending to be stored */
  int isctr; /* true if this is a constructor */
};


#define ccbase(cc) ((cc)->bl->basereg)

/* constructfields is mutually recursive with forclause */
static void forcomprehensionbody (LexState *ls, void *vcc);
/* We might be able to get rid of this one by juggling code order */
static void body (LexState *ls, expdesc *e, int needself, int line);


static void constructorset (FuncState *fs, struct ConsControl *cc,
                            int vecbase, int tostore) {
  luaK_codeABC(fs, OP_CALL, vecbase, tostore, 3);
  if (ccbase(cc) + 2 != vecbase)
    luaK_codeABC(fs, OP_MOVE, ccbase(cc) + 2, vecbase, 0);
  luaK_codeABC(fs, OP_TESTNIL, ccbase(cc) + 3, vecbase + 1, 0);
  luaK_concat(fs, &cc->bl->breaklist, luaK_jump(fs));
}


static void closelistfield (FuncState *fs, struct ConsControl *cc,
                            int vecbase, int multi) {
  int tostore = cc->tostore;
  if (tostore != 0) {
    lua_assert(cc->v.k != VVOID);
    if (multi && hasmultret(cc->v.k)) {
      luaK_setmultret(fs, &cc->v);
      tostore = 0;
      if (cc->na == 1) cc->na = 0;
    }
    else
      luaK_exp2nextreg(fs, &cc->v);
    if (cc->isctr)
      constructorset(fs, cc, vecbase, tostore ? tostore + 4 : 0);
    else
      luaK_codeABC(fs, OP_SETLIST, ccbase(cc), tostore, vecbase);
    fs->freereg = vecbase;
    cc->tostore = 0;
    cc->v.k = VVOID;
  }
}


static void recfield (LexState *ls, struct ConsControl *cc, int vecbase) {
  /* recfield -> (NAME | `['exp1`]') = exp1 */
  FuncState *fs = ls->fs;
  expdesc key, val;
  int doing_docstring = 0;
  closelistfield(fs, cc, vecbase, 0);
  if (cc->isctr) {
    luaK_codeABC(fs, OP_MOVE3, vecbase, ccbase(cc), 0);
    fs->freereg = vecbase + 3;
  }
  if (testnext(ls, TK_FUNCTION)) {
    int needself = testnext(ls, ':');
    checkname(ls, &key);
    if (cc->isctr)
      luaK_exp2nextreg(fs, &key);
    body(ls, &val, needself, ls->linenumber);
  }
  else {
    init_exp(&val, VVOID, 0);
    if (ls->t.token == TK_DOCSTRING) {
      TString *docstr_magic_key = luaS_newliteral(ls->L, "__aranha_docstring");
      init_exp(&key, VK, luaK_stringK(ls->fs, docstr_magic_key));
      init_exp(&val, VK, luaK_stringK(ls->fs, ls->t.seminfo.ts));
      luaX_next(ls);
      doing_docstring = 1;
    }
    else {
      if (ls->t.token == TK_NAME)
        checkname(ls, &key);
      else if (ls->t.token == '[')
        yindex(ls, &key);
      checknext(ls, '=');
    }
    if (cc->isctr)
      luaK_exp2nextreg(fs, &key);
    else
      luaK_exp2RK(fs, &key);
    if (!doing_docstring)
      expr(ls, &val);
  }
  cc->nh++;
  if (cc->isctr) {
    luaK_exp2nextreg(fs, &val);
    constructorset(fs, cc, vecbase, 5);
  }
   else
    luaK_codeABC(fs, OP_SETTABLE, ccbase(cc), luaK_exp2RK(fs, &key),
                                              luaK_exp2RK(fs, &val));
  fs->freereg = vecbase;  /* free registers */
}


static void listfield (LexState *ls, struct ConsControl *cc, int vecbase) {
  if (cc->tostore != 0) {
    lua_assert(cc->v.k != VVOID);
    luaK_exp2nextreg(ls->fs, &cc->v);
  }
  else if (cc->isctr) {
    FuncState *fs = ls->fs;
    luaK_codeABC(fs, OP_MOVE3, vecbase, ccbase(cc), 0);
    luaK_nil(fs, vecbase+3, 1);
    fs->freereg = vecbase + 4;
  }
  expr(ls, &cc->v);
  cc->na++;
  cc->tostore++;
}


/* FIXME To actually capture what's going on, we need to use two
 * scopes; one with the function/object internals, and the second one
 * with just the state. We could then accurately reflect the fact that
 * the state is temporarily relocated between the MOVE3 and the
 * CALL. Actually, that's not quite right either...  maybe we need
 * three scopes.
 */
static void constructfields (LexState *ls, struct ConsControl *cc, 
                             int terminator) {
  FuncState *fs = ls->fs;
  int needcomma = 0;
  int wascomma = 0;
  int vecbase = fs->freereg;
  cc->v.k = VVOID;
  if (cc->isctr && ccbase(cc) + 3 == vecbase) --vecbase;
  while (ls->t.token != terminator) {
    needcomma = 1;
    wascomma = 0;
    if (cc->tostore == LFIELDS_PER_FLUSH)
      closelistfield(fs, cc, vecbase, 0);
    switch(ls->t.token) {
      case TK_DOCSTRING: { /* Aranha documentation string */
        needcomma = 0;
        recfield(ls, cc, vecbase);
        break;
      }
      case TK_NAME: {  /* may be listfields or recfields */
        luaX_lookahead(ls);
        if (ls->lookahead.token != '=')  /* expression? */
          listfield(ls, cc, vecbase);
        else {
          recfield(ls, cc, vecbase);
        }
        break;
      }
      case '[': {  /* constructor_item -> recfield */
        recfield(ls, cc, vecbase);
        break;
      }
      case TK_FOR: { /* constructor_item -> comprehension */
        int line = ls->linenumber;
        needcomma = 0;
        closelistfield(fs, cc, vecbase, 1);
        luaX_next(ls);
        if (cc->isctr && ccbase(cc) == vecbase + 2)
          luaK_reserveregs(fs, 1);
        forclause(ls, line, 1, forcomprehensionbody, cc);
        check_match(ls, TK_END, TK_FOR, line);
        break;
      }
      case TK_FUNCTION: { /* might be a recfield */
        luaX_lookahead(ls);
        if (ls->t.token != '(') {
          needcomma = 0;
          recfield(ls, cc, vecbase);
        }
        else
          listfield(ls, cc, vecbase);
        break;
      }
      default: {  /* constructor_part -> listfield */
        listfield(ls, cc, vecbase);
        break;
      }
    }
    if (testnext(ls, ';'))
      closelistfield(fs, cc, vecbase, 1);
    else if (testnext(ls, ','))
      wascomma = 1;
    else if (ls->t.token == TK_NAME || ls->t.token == TK_FOR || ls->t.token == TK_FUNCTION)
      continue;
    else if (needcomma)
      break;
  }
  closelistfield(fs, cc, vecbase, !wascomma);
}


static void forcomprehensionbody (LexState *ls, void *vcc) {
  struct ConsControl *cc = vcc;
  checknextsemi(ls, "yield");
  constructfields(ls, cc, TK_END);
}


static void ccinit (struct ConsControl *cc, BlockCnt *bl) {
  cc->na = cc->nh = cc->tostore = 0;
  cc->bl = bl;
  init_exp(&cc->v, VVOID, 0);  /* no value (yet) */
}

  
static void constructor (LexState *ls, struct ConsControl *cc) {
  /* constructor -> ?? */
   int line = ls->linenumber;
  checknext(ls, '{');
  constructfields(ls, cc, '}');
  check_match(ls, '}', '{', line);
}

  
static void constructorcall (LexState *ls) {
  BlockCnt bl;
  struct ConsControl cc;
  FuncState *fs = ls->fs;
  int s = newscope(fs);
  ccinit(&cc, &bl);
  cc.isctr = 1;
  new_localvarliteral(fs, s, "(constructor)");
  new_localvarliteral(fs, s, "(constructor object)");
  new_localvarliteral(fs, s, "(constructor state)");
  enterblock(fs, &bl, 0);
  activatescope(fs, s);
  constructor(ls, &cc);
  leaveblock(fs);
}


static void constructordisplay (LexState *ls, expdesc *t) {
  BlockCnt bl;
  struct ConsControl cc;
  FuncState *fs = ls->fs;
  int s = newscope(fs);
  int pc;
  ccinit(&cc, &bl);
  cc.isctr = 0;
  pc = luaK_codeABC(fs, OP_NEWTABLE, 0, 0, 0);
  init_exp(t, VRELOCABLE, pc);
  new_localvarliteral(fs, s, "(table)");
  new_localvarliteral(fs, s, "(table index)");
  enterblock(fs, &bl, 0);
  luaK_exp2nextreg(ls->fs, t);  /* fix it at start of block */
  activatescope(fs, s);
  constructor(ls, &cc);
  SETARG_B(fs->f->code[pc], luaO_int2fb(cc.na)); /* set initial array size */
  SETARG_C(fs->f->code[pc], luaO_int2fb(cc.nh)); /* set initial table size */
  leaveblock(fs);
  luaK_reserveregs(ls->fs, 1);  /* table still on top */
}

/* }====================================================================== */



static int localvarlist (LexState *ls) {
  /* localvarlist -> name { `,' name } */ 
  FuncState *fs = ls->fs;
  int s = newscope(fs);
  do {
    new_localvar(fs, s, str_checkname(ls));
  } while (testnext(ls, ','));
  return s;
}


static int parlist (LexState *ls, int s) {
  /* parlist -> [ param { `,' param } ] */
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  int nparams;
  f->is_vararg = 0;
  if (ls->t.token != ')') {  /* is `parlist' not empty? */
    do {
      if (ls->t.token == TK_NAME)  /* param -> NAME */
        new_localvar(fs, s, str_checkname(ls));
      else if (ls->t.token == TK_DOTS) {  /* param -> `...' */
        luaX_next(ls);
#if defined(LUA_COMPAT_VARARG)
        /* use `arg' as default name */
        new_localvarliteral(fs, s, "arg");
        f->is_vararg = VARARG_HASARG | VARARG_NEEDSARG;
#endif
        f->is_vararg |= VARARG_ISVARARG;
        break;
      }
      else
        luaX_syntaxerror(ls, "<name> or " LUA_QL("...") " expected");
    } while (testnext(ls, ','));
  }
  nparams = scopesize(fs, s);
  f->numparams = nparams - (f->is_vararg & VARARG_HASARG);
  return nparams;
}


static void body (LexState *ls, expdesc *e, int needself, int line) {
  /* body ->  `(' parlist `)' chunk END */
  FuncState *fs = ls->fs;
  FuncState new_fs;
  int s;
  open_func(ls, &new_fs);
  s = newscope(&new_fs);
  new_fs.f->linedefined = line;
  checknext(ls, '(');
  if (needself)
    new_localvarliteral(&new_fs, s, "self");
  if (parlist(ls, s) > 0)
    activatescope(&new_fs, s);
  else
    new_fs.nscopes = 0;  /* Ditch the scope, we don't need it */
  checknext(ls, ')');
  if (ls->t.token == TK_DOCSTRING) {
    /* Capture this string as the function's documentation */
    new_fs.f->doc = ls->t.seminfo.ts;
    luaX_next(ls);
  }
  chunk(ls);
  new_fs.f->lastlinedefined = ls->linenumber;
  check_match(ls, TK_END, TK_FUNCTION, line);
  if (needself) {
    /* Rename initial variable to (self) which can't be done from within
     * Lua source code itself. This serves as a sentinel for method functions
     */
    new_fs.f->locvars[0] = luaX_newstring(ls, "(self)", 6);
  }
  close_func(ls);
  pushclosure(fs, &new_fs, e);
}


static int explist1 (LexState *ls, expdesc *v) {
  /* explist1 -> expr { `,' expr } */
  int n = 1;  /* at least one expression */
  expr(ls, v);
  while (testnext(ls, ',')) {
    luaK_exp2nextreg(ls->fs, v);
    expr(ls, v);
    n++;
  }
  return n;
}


static void assign_scope (LexState *ls, int s) {
  FuncState *fs = ls->fs;
  expdesc e;
  int nexps = explist1(ls, &e);
  adjust_assign(fs, scopesize(fs, s), nexps, &e);
  activatescope(fs, s);
}


static void nil_scope (LexState *ls, int s) {
  FuncState *fs = ls->fs;
  int extra = scopesize(fs, s);
  luaK_reserveregs(fs, extra);
  luaK_nil(fs, scopebase(fs, s), extra);
  activatescope(fs, s);
}


static void funcargs (LexState *ls, expdesc *f) {
  FuncState *fs = ls->fs;
  expdesc args;
  int base, nparams;
  int line = ls->linenumber;
  int calltype = OP_CALL;
  args.k = VVOID;
  switch (ls->t.token) {
    case '(': {  /* funcargs -> `(' [ explist1 ] `)' */
      if (line != ls->lastline)
        luaX_syntaxerror(ls,"ambiguous syntax (function call x new statement)");
      luaX_next(ls);
      if (testnext(ls, ')'))  /* arg list is empty? */
        break;
      else {
        explist1(ls, &args);
        luaK_setmultret(fs, &args);
      }
      check_match(ls, ')', '(', line);
      break;
    }
    case '{': {  /* funcargs -> constructor */
      calltype = OP_CTRCLOSE;
      break;
    }
    case TK_STRING: {  /* funcargs -> STRING */
      codestring(ls, &args, ls->t.seminfo.ts);
      luaX_next(ls);  /* must use `seminfo' before `next' */
      break;
    }
    default: {
      luaX_syntaxerror(ls, "function arguments expected");
      return;
    }
  }
  lua_assert(f->k == VNONRELOC);
  base = f->u.s.info;  /* base register for call */
  if (hasmultret(args.k))
    nparams = LUA_MULTRET;  /* open call */
  else {
    if (args.k != VVOID)
      luaK_exp2nextreg(fs, &args);  /* close last argument */
    nparams = fs->freereg - (base+1);
  }
  if (calltype == OP_CTRCLOSE) {
    lua_assert(nparams == 0 || nparams == 1);
    luaK_codeABC(fs, OP_CTROPEN, base, nparams+1, 0);
    constructorcall(ls);
    luaK_codeABC(fs, OP_CLOSE, base, 0, 0);  /* FIXME */
  }
  init_exp(f, VCALL, luaK_codeABC(fs, calltype, base, nparams+1, 2));
  luaK_fixline(fs, line);
  fs->freereg = base+1;  /* call removes function and arguments and leaves
                            (unless changed) one result */
}




/*
** {======================================================================
** Expression parsing
** =======================================================================
*/


static void prefixexp (LexState *ls, expdesc *v) {
  /* prefixexp -> NAME | '(' expr ')' */
  switch (ls->t.token) {
    case '(': {
      int line = ls->linenumber;
      luaX_next(ls);
      expr(ls, v);
      check_match(ls, ')', '(', line);
      luaK_dischargevars(ls->fs, v);
      return;
    }
    case TK_NAME: {
      singlevar(ls, v);
      return;
    }
    default: {
      luaX_syntaxerror(ls, "unexpected symbol");
      return;
    }
  }
}


static void primaryexp (LexState *ls, expdesc *v) {
  /* primaryexp ->
        prefixexp { `.' NAME | `[' exp `]' | `:' NAME funcargs | funcargs } */
  FuncState *fs = ls->fs;
  prefixexp(ls, v);
  for (;;) {
    switch (ls->t.token) {
      case '.': {  /* field */
        field(ls, v);
        break;
      }
      case '[': {  /* `[' exp1 `]' */
        expdesc key;
        luaK_exp2anyreg(fs, v);
        yindex(ls, &key);
        luaK_indexed(fs, v, &key);
        break;
      }
      case ':': {  /* `:' NAME funcargs */
        expdesc key;
        luaX_next(ls);
        checkname(ls, &key);
        luaK_self(fs, v, &key);
        funcargs(ls, v);
        break;
      }
      case '(': case TK_STRING: case '{': {  /* funcargs */
        luaK_exp2nextreg(fs, v);
        funcargs(ls, v);
        break;
      }
      default: return;
    }
  }
}


static void simpleexp (LexState *ls, expdesc *v) {
  /* simpleexp -> NUMBER | STRING | NIL | true | false | ... |
                  constructor | FUNCTION body | primaryexp */
  switch (ls->t.token) {
    case TK_NUMBER: {
      init_exp(v, VKNUM, 0);
      v->u.nval = ls->t.seminfo.r;
      break;
    }
    case TK_STRING: {
      codestring(ls, v, ls->t.seminfo.ts);
      break;
    }
    case TK_NIL: {
      init_exp(v, VNIL, 0);
      break;
    }
    case TK_TRUE: {
      init_exp(v, VTRUE, 0);
      break;
    }
    case TK_FALSE: {
      init_exp(v, VFALSE, 0);
      break;
    }
    case TK_DOTS: {  /* vararg */
      FuncState *fs = ls->fs;
      check_condition(ls, fs->f->is_vararg,
                      "cannot use " LUA_QL("...") " outside a vararg function");
      fs->f->is_vararg &= ~VARARG_NEEDSARG;  /* don't need 'arg' */
      init_exp(v, VVARARG, luaK_codeABC(fs, OP_VARARG, 0, 1, 0));
      break;
    }
    case '{': {  /* constructor */
      constructordisplay(ls, v);
      return;
    }
    case TK_FUNCTION: {
      luaX_next(ls);
      body(ls, v, testnext(ls, ':'), ls->linenumber);
      return;
    }
    default: {
      primaryexp(ls, v);
      return;
    }
  }
  luaX_next(ls);
}


static UnOpr getunopr (int op) {
  switch (op) {
    case TK_NOT: return OPR_NOT;
    case '-': return OPR_MINUS;
    case '#': return OPR_LEN;
    default: return OPR_NOUNOPR;
  }
}


static BinOpr getbinopr (int op) {
  switch (op) {
    case '+': return OPR_ADD;
    case '-': return OPR_SUB;
    case '*': return OPR_MUL;
    case '/': return OPR_DIV;
    case '%': return OPR_MOD;
    case '^': return OPR_POW;
    case TK_CONCAT: return OPR_CONCAT;
    case TK_NE: return OPR_NE;
    case TK_EQ: return OPR_EQ;
    case '<': return OPR_LT;
    case TK_LE: return OPR_LE;
    case '>': return OPR_GT;
    case TK_GE: return OPR_GE;
    case TK_AND: return OPR_AND;
    case TK_OR: return OPR_OR;
    default: return OPR_NOBINOPR;
  }
}


static const struct {
  lu_byte left;  /* left priority for each binary operator */
  lu_byte right; /* right priority */
} priority[] = {  /* ORDER OPR */
   {6, 6}, {6, 6}, {7, 7}, {7, 7}, {7, 7},  /* `+' `-' `/' `%' */
   {10, 9}, {5, 4},                 /* power and concat (right associative) */
   {3, 3}, {3, 3},                  /* equality and inequality */
   {3, 3}, {3, 3}, {3, 3}, {3, 3},  /* order */
   {2, 2}, {1, 1}                   /* logical (and/or) */
};

#define UNARY_PRIORITY	8  /* priority for unary operators */


/*
** subexpr -> (simpleexp | unop subexpr) { binop subexpr }
** where `binop' is any binary operator with a priority higher than `limit'
*/
static BinOpr subexpr (LexState *ls, expdesc *v, unsigned int limit) {
  BinOpr op;
  UnOpr uop;
  enterlevel(ls);
  uop = getunopr(ls->t.token);
  if (uop != OPR_NOUNOPR) {
    luaX_next(ls);
    subexpr(ls, v, UNARY_PRIORITY);
    luaK_prefix(ls->fs, uop, v);
  }
  else simpleexp(ls, v);
  /* expand while operators have priorities higher than `limit' */
  op = getbinopr(ls->t.token);
  while (op != OPR_NOBINOPR && priority[op].left > limit) {
    expdesc v2;
    BinOpr nextop;
    luaX_next(ls);
    luaK_infix(ls->fs, op, v);
    /* read sub-expression with higher priority */
    nextop = subexpr(ls, &v2, priority[op].right);
    luaK_posfix(ls->fs, op, v, &v2);
    op = nextop;
  }
  leavelevel(ls);
  return op;  /* return first untreated operator */
}


static void expr (LexState *ls, expdesc *v) {
  subexpr(ls, v, 0);
}

/* }==================================================================== */



/*
** {======================================================================
** Rules for Statements
** =======================================================================
*/


static int block_follow (int token) {
  switch (token) {
    case TK_ELSE: case TK_ELSEIF: case TK_END:
    case TK_UNTIL: case TK_EOS:
      return 1;
    default: return 0;
  }
}


static void block (LexState *ls) {
  /* block -> chunk */
  FuncState *fs = ls->fs;
  BlockCnt bl;
  enterblock(fs, &bl, 0);
  chunk(ls);
  lua_assert(bl.breaklist == NO_JUMP);
  leaveblock(fs);
}


/*
** structure to chain all variables in the left-hand side of an
** assignment
*/
struct LHS_assign {
  struct LHS_assign *prev;
  expdesc v;  /* variable (global, local, upvalue, or indexed) */
};


/*
** check whether, in an assignment to a local variable, the local variable
** is needed in a previous assignment (to a table). If so, save original
** local value in a safe place and use this safe copy in the previous
** assignment.
*/
static void check_conflict (LexState *ls, struct LHS_assign *lh, expdesc *v) {
  FuncState *fs = ls->fs;
  int extra = fs->freereg;  /* eventual position to save local variable */
  int conflict = 0;
  for (; lh; lh = lh->prev) {
    if (lh->v.k == VINDEXED) {
      if (lh->v.u.s.info == v->u.s.info) {  /* conflict? */
        conflict = 1;
        lh->v.u.s.info = extra;  /* previous assignment will use safe copy */
      }
      if (lh->v.u.s.aux == v->u.s.info) {  /* conflict? */
        conflict = 1;
        lh->v.u.s.aux = extra;  /* previous assignment will use safe copy */
      }
    }
  }
  if (conflict) {
    luaK_codeABC(fs, OP_MOVE, fs->freereg, v->u.s.info, 0);  /* make copy */
    luaK_reserveregs(fs, 1);
  }
}


static void assignment (LexState *ls, struct LHS_assign *lh, int nvars) {
  FuncState *fs = ls->fs;
  expdesc e;
  check_condition(ls, VLOCAL <= lh->v.k && lh->v.k <= VINDEXED,
                      "syntax error");
  if (testnext(ls, ',')) {  /* assignment -> `,' primaryexp assignment */
    struct LHS_assign nv;
    nv.prev = lh;
    primaryexp(ls, &nv.v);
    if (nv.v.k == VLOCAL)
      check_conflict(ls, lh, &nv.v);
    assignment(ls, &nv, nvars+1);
  }
  else {  /* assignment -> `=' explist1 */
    int nexps;
    checknext(ls, '=');
    nexps = explist1(ls, &e);
    if (nexps != nvars) {
      adjust_assign(fs, nvars, nexps, &e);
      if (nexps > nvars)
        fs->freereg -= nexps - nvars;  /* remove extra values */
    }
    else {
      luaK_setoneret(fs, &e);  /* close last expression */
      luaK_storevar(fs, &lh->v, &e);
      return;  /* avoid default */
    }
  }
  init_exp(&e, VNONRELOC, fs->freereg-1);  /* default assignment */
  luaK_storevar(fs, &lh->v, &e);
}


static void condbreak (LexState *ls, BlockCnt *targ) {
  FuncState *fs = ls->fs;
  expdesc v;
  expr(ls, &v);
  if (needsclose(fs, targ)) {
    if (v.k == VK) v.k = VTRUE;
    luaK_goiffalse(fs, &v);
    luaK_codeABC(fs, OP_CLOSE, targ->basereg, 0, 0);
    luaK_concat(fs, &targ->breaklist, luaK_jump(fs));
    luaK_patchtohere(fs, v.t);
  }
  else {
    if (v.k == VNIL) v.k = VFALSE;
    luaK_goiftrue(fs, &v);
    luaK_concat(fs, &targ->breaklist, v.f);
  }
}


static void breakstat (LexState *ls) {
  FuncState *fs = ls->fs;
  BlockCnt *bl = findbreakable(fs);
  if (!bl)
    luaX_syntaxerror(ls, "no loop to break");
  codebreak(fs, bl);
}


static void satisfies (LexState *ls) {
  /* satisifes -> (varlist = exprlist SATISFIES)? */
  if (ls->t.token == TK_NAME) {
    luaX_lookahead(ls);
    if (ls->lookahead.token == '=' || ls->lookahead.token == ',') {
      int s = localvarlist(ls);
      checknext(ls, '=');  /* This is why we couldn't just use localstat() */
      assign_scope(ls, s);
      checknextsemi(ls, "satisfies");
    }
  }
}


static void whilestat (LexState *ls, int line) {
  /* whilestat -> WHILE cond DO block END */
  FuncState *fs = ls->fs;
  int whileinit;
  BlockCnt bl1;
  BlockCnt bl2;
  luaX_next(ls);  /* skip WHILE */
  whileinit = luaK_getlabel(fs);
  enterblock(fs, &bl1, 1);
  enterblock(fs, &bl2, 0);
  satisfies(ls);
  condbreak(ls, &bl1);
  checknext(ls, TK_DO);
  block(ls);
  check_match(ls, TK_END, TK_WHILE, line);
  leaveblock(fs);
  luaK_patchlist(fs, luaK_jump(fs), whileinit);
  leaveblock(fs);
}


static void repeatstat (LexState *ls, int line) {
  /* repeatstat -> REPEAT block UNTIL cond */
  FuncState *fs = ls->fs;
  int repeat_init = luaK_getlabel(fs);
  BlockCnt bl1, bl2;
  enterblock(fs, &bl1, 1);  /* loop block */
  enterblock(fs, &bl2, 0);  /* scope block */
  luaX_next(ls);  /* skip REPEAT */
  chunk(ls);
  check_match(ls, TK_UNTIL, TK_REPEAT, line);
  condbreak(ls, &bl2);
  luaK_patchlist(ls->fs, bl2.breaklist, repeat_init);
  bl2.breaklist = NO_JUMP;
  leaveblock(fs);
  leaveblock(fs);  /* finish loop */
}


static int exp1 (LexState *ls) {
  expdesc e;
  int k;
  expr(ls, &e);
  k = e.k;
  luaK_exp2nextreg(ls->fs, &e);
  return k;
}


static int fornum (LexState *ls, int cs) {
  /* fornum -> NAME = exp1,exp1[,exp1] forbody */
  FuncState *fs = ls->fs;
  int preplist;
  new_localvarliteral(fs, cs, "(for index)");
  new_localvarliteral(fs, cs, "(for limit)");
  new_localvarliteral(fs, cs, "(for step)");
  exp1(ls);  /* initial value */
  checknext(ls, ',');
  exp1(ls);  /* limit */
  if (testnext(ls, ','))
    exp1(ls);  /* optional step */
  else {  /* default step = 1 */
    luaK_codeABx(fs, OP_LOADK, fs->freereg, luaK_numberK(fs, 1));
    luaK_reserveregs(fs, 1);
  }
  preplist = luaK_codeAsBx(fs, OP_FORPREP, scopebase(fs, cs), NO_JUMP);
  activatescope(fs, cs);
  return preplist;
}


static int forlist (LexState *ls, int cs) {
  /* forlist -> NAME {,NAME} IN explist1 forbody */
  FuncState *fs = ls->fs;
  /* create control variables */
  new_localvarliteral(fs, cs, "(for generator)");
  new_localvarliteral(fs, cs, "(for state)");
  new_localvarliteral(fs, cs, "(for control)");
  assign_scope(ls, cs);
  luaK_checkstack(fs, fs->freereg + 3);  /* extra space to call generator */
  return NO_JUMP;
}


static void shiftscopes (FuncState *fs, int basescope, int base, int offset) {
  int s;
  for (s = basescope; s < fs->nscopes; ++s) {
    if (scopebase(fs, s) >= base && isparentscope(fs, s, basescope)) {
      /* printf("Shifting scope %d by %d\n", s, offset); */
      getscope(fs, s)->slot += offset;
    }
  }
}


typedef struct ForCnt {
  BlockCnt outerbl;  /* The block for while's to break out of */
  BlockCnt innerbl;  /* The first inner block */
  int preplist;
  int condinit;  /* start of conds */
  int initinit;  /* start of inits */
  int condjpc;  /* jpc at end of conds */
  short initscope;  /* last scope for inits */
  short basescope;  /* first scope for loops */
} ForCnt;


static int savejpc (FuncState *fs) {
  int jpc = fs->jpc;
  fs->jpc = NO_JUMP;
  return jpc;
}


static void restorejpc (FuncState *fs, int *jpc) {
  fs->jpc = *jpc;
  *jpc = NO_JUMP;
}


static int forinit (LexState *ls, ForCnt *fc, int cs, int s) {
  /* forinit -> ('=' fornum | IN forlist) */
  FuncState *fs = ls->fs;
  int isnum = 0;
  fc->initinit = luaK_getlabel(fs);
  switch (ls->t.token) {
    case '=': {
      if (scopesize(fs, s) > 1) error_expected(ls, TK_IN);
      luaX_next(ls);
      fc->preplist = fornum(ls, cs);
      isnum = 1;
      break;
    }
    case TK_IN: {
      luaX_next(ls);
      fc->preplist = forlist(ls, cs);
      break;
    }
    default:
      luaX_syntaxerror(ls, LUA_QL("=") " or " LUA_QL("in") " expected");
  }
  fc->initscope = cs;
  return isnum;
}
  

static void forconds (LexState *ls, ForCnt *fc) {
  /* forconds = (WHERE | WHILE) [satisfies] condition */
  FuncState *fs = ls->fs;
  BlockCnt *targ;
  for (;;) {
    if (testnext(ls, TK_WHILE))
      targ = &fc->outerbl;
    else if (testnextsemi(ls, "when"))
      targ = fs->bl;
    else break;
    satisfies(ls);
    condbreak(ls, targ);
  }
  fc->condjpc = savejpc(fs);
}

    
static int codeforloop(FuncState *fs, int isnum, int line, int cs, int s) {
  int jumplist, oldline;
  int cbase = scopebase(fs, cs);
  int vbase = scopebase(fs, s);
  oldline = fs->ls->lastline;
  fs->ls->lastline = line;
  if (isnum)
    luaK_codeABC(fs, OP_FORLOOP, vbase, cbase, 0);
  else {
    luaK_codeABC(fs, OP_TFORCALL, vbase, cbase, scopesize(fs, s) + 1);
    luaK_codeABC(fs, OP_TESTNIL, cbase + 2, vbase, 0);
  }
  jumplist = luaK_jump(fs);
  fs->ls->lastline = line;
  return jumplist;
}

static void fixlist (FuncState *fs, int *list, int part1, int part2) {
  if (*list != NO_JUMP && *list >= part1) {
    if (*list >= part2)
      *list -= part2 - part1;
    else if (*list < fs->pc)
      *list += (fs->pc - part2);
  }
}


/* void PrintRange(const Proto* f, int st, int fi); */

/* FIXME: This will occasionally produce a JUMP to a JUMP */
static void rotatefor (FuncState *fs, ForCnt *fc) {
  /*  printf("Rotating %d->%d pc = %d\n", fc->condinit, fc->initinit, fs->pc); */
  if (fc->condinit != fc->initinit) {
    int jpc = savejpc(fs);
    /*
       printf("Before\n");
       PrintRange(fs->f, 0, fs->pc);
    */
    luaCU_rotatecode(fs, fc->condinit, fc->initinit - fc->condinit);
    fixlist(fs, &fc->outerbl.breaklist, fc->condinit, fc->initinit);
    fixlist(fs, &fc->innerbl.breaklist, fc->condinit, fc->initinit);
    fixlist(fs, &fc->preplist, fc->condinit, fc->initinit);
    fixlist(fs, &fc->condjpc, fc->condinit, fc->initinit);
    fixlist(fs, &jpc, fc->condinit, fc->initinit);
    luaK_patchlist(fs, jpc, fc->condinit + fs->pc - fc->initinit);
    /*
       printf("After\n");
       PrintRange(fs->f, 0, fs->pc);
    */
  }
  restorejpc(fs, &fc->condjpc);
  /* printf("jpc: %d\n", fs->jpc); */
}


/* We're in the innerblock when we execute this. That's actually
 * semantically incorrect, but it works anyway. This all needs to be
 * rethought according to the new BlockCnt structures.
 */
static void andforclause (LexState *ls, ForCnt *fc) {
  FuncState *fs = ls->fs;
  BlockCnt bl;
  int endfor, loopinit;
  int olds = closescopes(fs, fc->initscope);
  int s = localvarlist(ls);
  int cs = newscope(fs);
  int preplist = fc->preplist;
  int isnum = forinit(ls, fc, cs, s);
  int line = ls->linenumber;
  /* Inject new init scope into scope chain */
  /* FIXME: Move the next five lines into codeutils,
   *        plus shiftscopes and isparentscope
   * When we do that, get rid of fc->basescope; we can
   * derive it when we're tracking down to initscope
  */
  int offset = scopesize(fs, cs);
  getscope(fs, fc->basescope)->outer = cs;
  luaCU_shiftcode(fs, fc->condinit, fc->initinit,
                  fc->innerbl.basereg, offset);
  shiftscopes(fs, fc->basescope, fc->innerbl.basereg, offset);
  fc->innerbl.basereg += offset;
  /* Rotate code */
  rotatefor(fs, fc);
  /* condinit is where the new init code ends up */
  luaK_patchlist(fs, preplist, fc->condinit);
  fc->condinit += fs->pc - fc->initinit;
  
  enterblock(fs, &bl, 0);
  reactivatescope(fs, olds);
  reanchorscope(fs, s);
  activatescope(fs, s);
  loopinit = luaK_getlabel(fs);
  endfor = codeforloop(fs, isnum, line, cs, s);
  codebreak(fs, &fc->outerbl);
  luaK_patchtohere(fs, endfor);
  forconds(ls, fc);
  luaK_patchlist(fs, bl.breaklist, loopinit);
  mergeblock(fs);
}


static void forclause (LexState *ls, int line, int breakable,
                       void(*bodyfunc)(LexState *ls, void *ud), void *ud) {
  /* forstat -> FOR (fornum | forlist) forbody END */
  FuncState *fs = ls->fs;
  ForCnt fc;
  int isnum;
  int s = localvarlist(ls);
  int cs = newscope(fs);
  fc.preplist = fc.condjpc = NO_JUMP;
  fc.basescope = s;
  enterblock(fs, &fc.outerbl, breakable);  /* scope for loop and control variables */
  isnum = forinit(ls, &fc, cs, s);
  enterblock(fs, &fc.innerbl, 0);
  reanchorscope(fs, s);
  activatescope(fs, s);
  fc.condinit = luaK_getlabel(fs);
  forconds(ls, &fc);
  while (testnextsemi(ls, "andfor")) {
    andforclause(ls, &fc);
  }
  if (fc.preplist == NO_JUMP) {
    fc.initinit = luaK_getlabel(fs);
    fc.preplist = luaK_jump(fs);
    rotatefor(fs, &fc);
  }
  else
    restorejpc(fs, &fc.condjpc);
  while (testnextsemi(ls, "with")) {
    int s = localvarlist(ls);
    checknext(ls, '=');
    assign_scope(ls, s);
    forconds(ls, &fc);
  }
  if (testnext(ls, TK_FOR))
    forclause(ls, ls->linenumber, 0, bodyfunc, ud);
  else 
    bodyfunc(ls, ud);
  leaveblock(fs);
  luaK_patchtohere(fs, fc.preplist);
  luaK_patchlist(fs, codeforloop(fs, isnum, line, cs, s), fc.preplist + 1);
  leaveblock(fs);
} 


static void test_then_block (LexState *ls, BlockCnt *bl) {
  /* test_then_block -> [IF | ELSEIF] cond THEN block */
  luaX_next(ls);  /* skip IF or ELSEIF */
  satisfies(ls);
  condbreak(ls, bl);
  checknext(ls, TK_THEN);
  chunk(ls);  /* `then' part */
}


static void ifstat (LexState *ls, int line) {
  /* ifstat -> IF cond THEN block {ELSEIF cond THEN block} [ELSE block] END */
  FuncState *fs = ls->fs;
  BlockCnt outerbl;
  BlockCnt innerbl;
  enterblock(fs, &outerbl, 0);
  enterblock(fs, &innerbl, 0);
  test_then_block(ls, &innerbl);  /* IF cond THEN block */
  while (ls->t.token == TK_ELSEIF) {
    codebreak(fs, &outerbl);
    leaveblock(fs);
    enterblock(fs, &innerbl, 0);
    test_then_block(ls, &innerbl);  /* ELSEIF cond THEN block */
  }
  if (ls->t.token == TK_ELSE) {
    codebreak(fs, &outerbl);
    leaveblock(fs);
    luaX_next(ls);  /* skip ELSE */
    block(ls);  /* `else' part */
  }
  else
    leaveblock(fs);
  leaveblock(fs);
  check_match(ls, TK_END, TK_IF, line);
}


static void localfunc (LexState *ls) {
  expdesc v, b;
  FuncState *fs = ls->fs;
  int s = newscope(fs);
  int needself = ls->t.token == ':';
  if (needself) luaX_next(ls);
  new_localvar(fs, s, str_checkname(ls));
  init_exp(&v, VLOCAL, fs->freereg);
  activatescope(fs, s);
  body(ls, &b, needself, ls->linenumber);
  luaK_storevar(fs, &v, &b);
}


static void localstat (LexState *ls) {
  /* stat -> LOCAL NAME {`,' NAME} [`=' explist1] */
  int s = localvarlist(ls);
  if (testnext(ls, '='))
    assign_scope(ls, s);
  else 
    nil_scope(ls, s);
}


static int funcname (LexState *ls, expdesc *v) {
  /* funcname -> NAME {field} [`:' NAME] */
  int needself = 0;
  if (ls->t.token == ':') {
    needself = 1;
    luaX_next(ls); /* skip the leading : */
  }
  singlevar(ls, v);
  while (ls->t.token == '.')
    field(ls, v);
  if (ls->t.token == ':') {
    needself = 1;
    field(ls, v);
  }
  return needself;
}


static void funcstat (LexState *ls, int line) {
  /* funcstat -> FUNCTION funcname body */
  FuncState *fs = ls->fs;
  int needself;
  expdesc v, b;
  luaX_next(ls);  /* skip FUNCTION */
  needself = funcname(ls, &v);
  body(ls, &b, needself, line);
  luaK_storevar(fs, &v, &b);
  luaK_fixline(fs, line);  /* definition `happens' in the first line */
}


static void exprstat (LexState *ls) {
  /* stat -> func | assignment */
  FuncState *fs = ls->fs;
  struct LHS_assign v;
  primaryexp(ls, &v.v);
  if (v.v.k == VCALL)  /* stat -> func */
    SETARG_C(getcode(fs, &v.v), 1);  /* call statement uses no results */
  else {  /* stat -> assignment */
    v.prev = NULL;
    assignment(ls, &v, 1);
  }
}


static void retstat (LexState *ls) {
  /* stat -> RETURN explist */
  FuncState *fs = ls->fs;
  expdesc e;
  int first, nret;  /* registers with returned values */
  luaX_next(ls);  /* skip RETURN */
  if (block_follow(ls->t.token) || ls->t.token == ';')
    first = nret = 0;  /* return no values */
  else {
    first = fs->tempreg;
    nret = explist1(ls, &e);  /* optional return values */
    if (hasmultret(e.k)) {
      luaK_setmultret(fs, &e);
      if (e.k == VCALL && nret == 1) {  /* tail call? */
        if (GET_OPCODE(getcode(fs,&e)) == OP_CALL)
          SET_OPCODE(getcode(fs,&e), OP_TAILCALL);
        lua_assert(GETARG_A(getcode(fs,&e)) == first);
      }
      nret = LUA_MULTRET;  /* return all values */
    }
    else {
      if (nret == 1)  /* only one single value? */
        first = luaK_exp2anyreg(fs, &e);
      else {
        luaK_exp2nextreg(fs, &e);  /* values must go to the `stack' */
        lua_assert(nret == fs->freereg - first);
      }
    }
  }
  luaK_ret(fs, first, nret);
}


static void forchunkbody (LexState *ls, void *unused) {
  checknext(ls, TK_DO);
  chunk(ls);
}

/* We're borrowing these from lauxlib.c */

typedef struct DivertedLoadS {
  const char *s;
  size_t size;
} DivertedLoadS;


static const char *divertedgetS (lua_State *L, void *ud, size_t *size) {
  DivertedLoadS *ls = (DivertedLoadS *)ud;
  (void)L;
  if (ls->size == 0) return NULL;
  *size = ls->size;
  ls->size = 0;
  return ls->s;
}

/* This delimits the expressions in diverted strings. It should be
 * unlikely to appear in HTML or SQL since those are the most
 * common languages which will use it
 *
 * Since | is the current chosen delimiter there is the added cuteness
 * that two of DIVERTED_EXPRESSION_DELIMITER in a row will be considered
 * as part of the string rather than a reason to switch to expression mode.
 *
 * Also, DIVERTED_EXPRESSION_FORMAT_DELIMITER is used between |expr and | to
 * allow you to control the format of the value. E.g. |pi#1.4f|
 *
 * That optimisation is not passed on in the decision to use vdprint or not
 * so you still get the slightly slower __divertedstring call even if a
 * call to __verbatimdivertedstring would have done the job.
 */
#define DIVERTED_EXPRESSION_DELIMITER '|'
#define DIVERTED_EXPRESSION_FORMAT_DELIMITER '$'

#define DIVERTED_EXPRESSION_ACCUMULATE_CHAR(L,b,c) \
        if (b->n + 1 > b->buffsize) { \
          size_t newsize; \
          if (b->buffsize >= MAX_SIZET/2) \
            luaX_lexerror(ls, "lexical element too long", 0); \
          newsize = (b->buffsize * 2)+1; \
          luaZ_resizebuffer(L, b, newsize); \
        } \
        b->buffer[b->n++] = cast(char, c)


static void divertedstringstat (LexState *ls) {
  /* stat -> divertedstring */
  int f = 0, i = 0;
  const char* ts = getstr(ls->t.seminfo.ts);
  
  if( ls->t.seminfo.ts->tsv.len == 0 ) { 
    /* String of zero bytes is useless to us */
    luaX_next(ls);
    return;
  }
  /* We can't be sure a string is NULL terminated; so... */
  for (i = 0; i < ls->t.seminfo.ts->tsv.len; ++i, ++ts)
    if (*ts == DIVERTED_EXPRESSION_DELIMITER) { f = 1; break; }
  if( f == 0 ) {
    /* Basic case: verbatim string with no interpolation; process simply */
    int base;
    expdesc func, text;
    int line = ls->linenumber;
    TString *verb_dstring = luaS_newliteral(ls->L, "__verbatimdivertedstring");
    ls->lastline = line;
    /* Remember the registers */
    base = ls->fs->freereg;
    /* 1 load __divertedstring into a register */
    if (singlevaraux(ls->fs, verb_dstring, &func, 1) == VGLOBAL)
      func.u.s.info = luaK_stringK(ls->fs, verb_dstring);
    luaK_exp2nextreg(ls->fs, &func);
    /* 2 load the diverted string itself into a register */
    codestring(ls, &text, ls->t.seminfo.ts);
    luaK_exp2nextreg(ls->fs, &text);
    /* 3 make the function call, 1 arg, no returns */
    init_exp( &func, VCALL, luaK_codeABC(ls->fs, OP_CALL, base, 2, 1));
    luaK_fixline(ls->fs, line);
    ls->fs->freereg = base;
    /* 4. Move on. */
    luaX_next(ls);
    return;
  }
  /* More complex case of non-verbatim string needs a little more work
   * because we essentially have to construct a new lexer using a ZIO over
   * the TString of ls->t.seminfo.ts and get on with running it
   * over our string, We parse carefully until we hit a | and then we do
   * something odd with parsing an expression into a register somehow
   */
  {
    LexState newls;
    TString *newname;
    char buffer[2];
    int bufflen = 0;
    ZIO newz;
    DivertedLoadS str;
    int base = 0;
    expdesc func, temp;
    int line = ls->linenumber;
    int accumargs = 0;
    int regforformatstring;
    TString *dstring = luaS_newliteral(ls->L, "__divertedstring");
    
    ls->lastline = line;
    
    {
      char *namebuff;
      namebuff = luaM_malloc(ls->L, 128+ls->source->tsv.len);
      snprintf(namebuff, 128, "Diverted string around line %d of ", 
               ls->linenumber);
      namebuff[strlen(namebuff)+ls->source->tsv.len] = 0;
      memcpy(namebuff+strlen(namebuff), getstr(ls->source),
             ls->source->tsv.len);
      newname = luaS_new(ls->L, namebuff);
      luaM_freemem(ls->L, namebuff, 128+ls->source->tsv.len);
    }
    /* We now have a name for our new lexer... 
     *
     * First up, we need to preserve any contents in the buffer which we
     * are about to lend to our new lex state...
     *
     * Since we're being called due to a TK_DIVERTEDSTRING, there should
     * be *AT MOST* two characters in the buffer, if there are more then
     * we're fucked...
     *
     */
    if (luaZ_bufflen(ls->buff) > 2) {
      /* Throw an error because we cannot compile this */
      luaX_syntaxerror(ls,
                       luaO_pushfstring(ls->L, 
                                        "Buffer too large when parsing "\
                                        "diverted string. Expected 0,1,2 "\
                                        
                                        "chars, got %d instead.", 
                                        luaZ_bufflen(ls->buff)));
    }
    /* Okay, we can compile this, copy the buffer length and chars into
     * our temporary storage...
     */
    bufflen = luaZ_bufflen(ls->buff);
    if (bufflen > 0) {
      buffer[0] = luaZ_buffer(ls->buff)[0];
      if (bufflen > 1)
        buffer[1] = luaZ_buffer(ls->buff)[1];
    }
    luaZ_bufflen(ls->buff) = 0;
    luaZ_bufflen(ls->dbuff) = 0; /* Empty out the dbuffer */
    /* Now construct the ZIO over our diverted string */
    str.s = getstr(ls->t.seminfo.ts);
    str.size = ls->t.seminfo.ts->tsv.len;
    luaZ_init(ls->L, &newz, &divertedgetS, &str);
    /* Feed in the buffer we're using */
    newls.buff = ls->buff;
    newls.dbuff = ls->dbuff;
    /* And now the lexer over this new ZIO */
    luaX_setinput(ls->L, &newls, &newz, newname);
    /* Copy in the function state */
    newls.fs = ls->fs;
    /* Remember the registers */
    base = ls->fs->freereg;
    /* load __divertedstring into a register */
    if (singlevaraux(ls->fs, dstring, &func, 1) == VGLOBAL)
      func.u.s.info = luaK_stringK(ls->fs, dstring);
    luaK_exp2nextreg(ls->fs, &func);
    
    regforformatstring = ls->fs->freereg;
    luaK_reserveregs(ls->fs, 1); /* Reserve one register for the format str */
    
    luaZ_resetbuffer(newls.buff); /* We accumulate non-expressions here */
    while (newls.t.token != TK_EOS && newls.current != EOZ) {
      /* We're not going to have a lookahead because that's rude
       * Simply read the next byte and accumulate if not a pipe
       */
      if (newls.current != DIVERTED_EXPRESSION_DELIMITER ||
          (newls.current == DIVERTED_EXPRESSION_DELIMITER &&
           luaZ_lookahead(newls.z) == DIVERTED_EXPRESSION_DELIMITER)) {
        /* We want to accumulate this into the buffer, this means
         * we have to resize the buffer if it's too small
         * This is save() from llex.c because it's static
         */
        DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                            newls.current);
        if (newls.current == DIVERTED_EXPRESSION_DELIMITER) {
          newls.current = zgetc(newls.z); /* next(newls) from llex.c */
        }
        if (newls.current == '%') {
          DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                              newls.current);
        }
        newls.current = zgetc(newls.z); /* next(newls) from llex.c */
      } else {
        /* Consume the pipe */
        newls.current = zgetc(newls.z); /* next(newls) from llex.c */
	luaX_next(&newls);
        expr(&newls, &temp);
        luaK_exp2nextreg(ls->fs, &temp);
        accumargs++;
        /* Emit a % to the dbuffer */
        DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                            '%');
        /* The terminating pipe should have been absorbed into t.token */
        if (newls.t.token == DIVERTED_EXPRESSION_DELIMITER) {
          /* Okay, we haven't got a format string, emit an 's' */
          DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                              's');
        } else {
          if (newls.t.token == DIVERTED_EXPRESSION_FORMAT_DELIMITER) {
            /* Righty, we got what we expected, transfer until delimiter */
            while (newls.current != DIVERTED_EXPRESSION_DELIMITER) {
              DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff,
                                                  newls.current);
              newls.current = zgetc(newls.z); /* next(ls) from llex.c */
            }
            newls.current = zgetc(newls.z); /* Consume the delimiter */
          } else {
            /* Dunno what it is, abort abort abort */
            luaX_syntaxerror(ls,
                             luaO_pushfstring(ls->L, 
                                              "Expected `%c' or `%c', "\
                                              "got `%s'  "\
                                              "while parsing " \
                                              "diverted string.",
                                              DIVERTED_EXPRESSION_DELIMITER,
                                              DIVERTED_EXPRESSION_FORMAT_DELIMITER,
                                              luaX_token2str(&newls,
                                                             newls.t.token)
                                              ));
          }
        }
      }
    }
    {
      int regof;
      codestring(ls, &temp, luaX_newstring(&newls, newls.dbuff->buffer,
                                           newls.dbuff->n));
      regof = ls->fs->freereg;
      luaK_exp2nextreg(ls->fs, &temp);
      luaZ_resetbuffer(newls.dbuff);
      luaK_codeABC(ls->fs, OP_MOVE, regforformatstring, regof, 0);
      ls->fs->freereg = regof;
    }
    /* Finally, code up the VCALL */
    init_exp( &func, VCALL, luaK_codeABC(ls->fs, OP_CALL, base, accumargs+2, 1));
    luaK_fixline(ls->fs, line);
    ls->fs->freereg = base;
    /* We restore the content of our old lex buffer */
    if (luaZ_sizebuffer(ls->buff) < LUA_MINBUFFER)
      luaZ_resizebuffer(ls->L, ls->buff, LUA_MINBUFFER /* at least 2 */);
    luaZ_bufflen(ls->buff) = bufflen;
    if (bufflen > 0) {
      luaZ_buffer(ls->buff)[0] = buffer[0];
      if (bufflen > 1)
        luaZ_buffer(ls->buff)[1] = buffer[1];
    }
    /* Finally, consume the lexeme from our primary lexer */
    luaX_next(ls);
  }
}


static int statement (LexState *ls) {
  int line = ls->linenumber;  /* may be needed for error messages */
  switch (ls->t.token) {
    case TK_IF: {  /* stat -> ifstat */
      ifstat(ls, line);
      return 0;
    }
    case TK_WHILE: {  /* stat -> whilestat */
      whilestat(ls, line);
      return 0;
    }
    case TK_DO: {  /* stat -> DO block END */
      luaX_next(ls);  /* skip DO */
      block(ls);
      check_match(ls, TK_END, TK_DO, line);
      return 0;
    }
    case TK_FOR: {  /* stat -> forstat */
      luaX_next(ls);  /* skip FOR */
      forclause(ls, line, 1, forchunkbody, NULL);
      check_match(ls, TK_END, TK_FOR, line);
      return 0;
    }
    case TK_REPEAT: {  /* stat -> repeatstat */
      repeatstat(ls, line);
      return 0;
    }
    case TK_FUNCTION: {
      funcstat(ls, line);  /* stat -> funcstat */
      return 0;
    }
    case TK_LOCAL: {  /* stat -> localstat */
      luaX_next(ls);  /* skip LOCAL */
      if (testnext(ls, TK_FUNCTION))  /* local function? */
        localfunc(ls);
      else
        localstat(ls);
      return 0;
    }
    case TK_RETURN: {  /* stat -> retstat */
      retstat(ls);
      return 1;  /* must be last statement */
    }
    case TK_BREAK: {  /* stat -> breakstat */
      luaX_next(ls);  /* skip BREAK */
      breakstat(ls);
      return 1;  /* must be last statement */
    }
    case TK_DIVERTEDSTRING: { /* stat -> divertedstringstat */
      divertedstringstat(ls);
      return 0; /* Can be anywhere in a function */
    }
    default: {
      exprstat(ls);
      return 0;  /* to avoid warnings */
    }
  }
}


static void chunk (LexState *ls) {
  /* chunk -> { stat [`;'] } */
  int islast = 0;
  enterlevel(ls);
  while (!islast && !block_follow(ls->t.token)) {
    islast = statement(ls);
    testnext(ls, ';');
    lua_assert(ls->fs->f->maxstacksize >= ls->fs->freereg);
    lua_assert(ls->fs->freereg >= ls->fs->tempreg);
    ls->fs->freereg = ls->fs->tempreg;
  }
  leavelevel(ls);
}

/* }====================================================================== */
